<?php

namespace wisdmlabs\cpb;

/**
* This class handles the product search for addon product
*/
class JsonProductSearch
{
    public function __construct()
    {
        if (is_admin()) {
            /*
            * Ajax for the Search bar for products
            */
            add_action('wp_ajax_woocommerce_cpb_json_search_products_and_variations', array($this, 'jsonSearchProductsAndVariations'));
        }
    }

    /**
    * This function is used to Search the Products in the Search bar.
    */
    public function jsonSearchProductsAndVariations()
    {
        //We won't write at the end of this because jsonSearchProducts will automatically do that for us.
        self::jsonSearchProducts('', array('product', 'product_variation'));
    }

    /**
     * Search for products and echo json.
     * It is very similar to WooCommerce's WC_AJAX::jsonSearchProducts. But it supports WPML too.
     * Gets the products which are published having title similar to the searched term and they must not be previously included in the same Add-on list.
     *
     * @param string $term       (default: '')
     * @param string $post_types (default: array('product'))
     */
    public static function jsonSearchProducts($term = '', $post_types = array('product'))
    {
        global $wpdb;

        ob_start();

        check_ajax_referer('search-products', 'security');


        $term = self::getTerm($term);

        $like_term = '%'.$wpdb->esc_like($term).'%';

        $query = $wpdb->prepare("
            SELECT ID FROM {$wpdb->posts} posts LEFT JOIN {$wpdb->postmeta} postmeta ON posts.ID = postmeta.post_id
            WHERE posts.post_status = 'publish'
            AND (
                posts.post_title LIKE %s
                or posts.post_content LIKE %s
                OR (
                    postmeta.meta_key = '_sku' AND postmeta.meta_value LIKE %s
                )
            )
        ", $like_term, $like_term, $like_term);

        $query .= " AND posts.post_type IN ('".implode("','", array_map('esc_sql', $post_types))."')";

        $query .= self::getQueryPart();

        $posts = array_unique($wpdb->get_col($query));

        $excludedProducts = self::getExcludedProducts();

        $found_products = array();

        if (!empty($posts)) {
            foreach ($posts as $post) {
                $product = wc_get_product($post);

                if (!current_user_can('read_product', $post)) {
                    continue;
                }
                if (empty($product)) {
                    continue;
                }
                $product_type = $product->get_type();

                switch ($product_type) {
                    case 'simple':
                        $found_products = self::getSimpleProduct($post, $product, $found_products, $excludedProducts);
                        break;
                    case 'variable':
                        $found_products = self::addAllVariationsInSearchResults($post, $product, $found_products, $excludedProducts);
                        break;
                    case 'variation':
                        if (!$product || ($product->is_type('variation'))) {
	                        continue 2;
                        }
                        $found_products = self::getVariationProduct($post, $product, $found_products, $excludedProducts);
                        break;
                    default:
                        break;
                }
            }
        }
        $found_products = apply_filters('wdm_cpb_json_search_found_products', $found_products);
        wp_send_json($found_products);
    }


    /**
    * This function gets the simple product details.
    * @param array $post post having the title similar to searched term
    * @param array $product above posts having post-type product
    * @param array $found_products empty at first
    * @param array $excludedProducts products included previously.
    * @return array $found_products details of simple product having the similar title as searched .
    */
    public static function getSimpleProduct($post, $product, $found_products, $excludedProducts = array())
    {
        $img_url = '';
        if (!$img_url || $img_url == '') {
            $img_url = wp_get_attachment_url(get_post_thumbnail_id($post));
        }
        if (!$img_url || $img_url == '') {
            $img_url = WC()->plugin_url().'/assets/images/placeholder.png';
        }
        if (!in_array(md5($post), $excludedProducts)) {
            $max_val =  0;
            $bundle_item_quantity = get_post_meta($post, '_stock', true);
            if (!empty($bundle_item_quantity)) {
                $max_val = $bundle_item_quantity;
            }

            $sku =  $product->get_sku();

            $found_products[ md5($post) ] = array(
                'product_id' => $post,
                'min_val' => 0,
                'max_val' => $max_val,
                'item_quantity' => $bundle_item_quantity,
                'bundle_quantity' => 0,
                'product_type' => 'simple',
                'text_name' => rawurldecode($product->get_formatted_name()),
                'price' => $product->get_price(),
                'product_title' => $product->get_title(),
                'url' => admin_url("/post.php?post={$post}&action=edit"),
                'permalink' => get_permalink($post),
                'product_image' => $img_url,
                'sku' => $sku,
                );
        }
        return $found_products;
    }

    /**
    * This function gets the simple product details.
    * @param array $post post having the title similar to searched term
    * @param array $product above posts having post-type product
    * @param array $found_products empty at first
    * @param array $excludedProducts products included previously.
    * @return array $found_products details of simple product having the similar title as searched .
    */
    public static function getVariableProduct($post, $product, $found_products, $excludedProducts = array())
    {
        $img_url = '';
        if (!$img_url || $img_url == '') {
            $img_url = wp_get_attachment_url(get_post_thumbnail_id($post));
        }
        if (!$img_url || $img_url == '') {
            $img_url = WC()->plugin_url().'/assets/images/placeholder.png';
        }
        if (!in_array(md5($post), $excludedProducts)) {
            $max_val =  0;
            $bundle_item_quantity = get_post_meta($post, '_stock', true);
            if (!empty($bundle_item_quantity)) {
                $max_val = $bundle_item_quantity;
            }

            $sku = $product->get_sku();
            $text_name = !empty($sku) ? $product->get_title()." (".$sku.")" : $product->get_title()." (#".$post.")";
            $excluded_products = array(md5($post));
            $found_products[md5($post)] = array(
                'product_id' => $post,
                'min_val' => 0,
                'max_val' => $max_val,
                'item_quantity' => $bundle_item_quantity,
                'bundle_quantity' => 0,
                'product_type' => $product->get_type(),
                'price' => $product->get_price(),
                'text_name' => $text_name,
                'product_title' => $product->get_title(),
                'url' => admin_url("/post.php?post={$post}&action=edit"),
                'permalink' => get_permalink($post),
                'product_image' => $img_url,
                'sku' => $product->get_sku(),
                'childrens' => self::addAllVariationsInSearchResults($post, $product, array(), $excluded_products)
            );
        }
        return $found_products;
    }

    /**
    * This function is to fetch the products with specific variations in the search bar.
    * If variations are not empty fetch the variable product corresponding to variation for search bar.
    * @param array $post post having the title similar to searched term
    * @param array $product above posts having post-type product
    * @param array $found_products empty at first
    * @param array $excludedProducts products included previously.
    * @return array $found_products details of variation product having the similar title as searched .
    */
    public static function getVariationProduct($post, $product, $found_products, $excludedProducts)
    {
        if (version_compare(WC_VERSION, '3.0.0', '<')) {
            add_filter('woocommerce_get_product_attributes', 'cpbModifyProductAttributeNames', 99, 1);
            $attributesArray = $product->get_variation_attributes();
            remove_filter('woocommerce_get_product_attributes', 'cpbModifyProductAttributeNames', 99, 1);
        } else {
            add_filter('woocommerce_product_get_attributes', 'cpbModifyProductAttributeNames', 99, 1);
            $attributesArray = $product->get_attributes();
            remove_filter('woocommerce_product_get_attributes', 'cpbModifyProductAttributeNames', 99, 1);
        }
        
        // $variation_data = $product->get_variation_attributes();
        $variation_data = wc_get_product_variation_attributes($post);

        $skipVariation = self::getSkipVariationStatus($variation_data);

        if (!empty($skipVariation)) {
            $product->variation_data = $variation_data;
            $found_products = self::addVariationInSearchResults($product, $post, $variation_data, $found_products, $excludedProducts);
        }
        return $found_products;
    }

    /**
    * This function is to return the variation those are empty.
    * @param array $variation_data Variation data for variable products
    * @return boolean $skipVariation true if variation is empty.
    */
    public static function getSkipVariationStatus($variation_data)
    {
        $skipVariation = false;
        foreach ($variation_data as $variation) {
            if (empty($variation)) {
                $skipVariation = true;
                break;
            }
        }

        return $skipVariation;
    }


    /**
    * Get the term similar to Product in search bar.
    * @param string $term term in search bar
    */
    public static function getTerm($term)
    {
        if (empty($term)) {
            $term = wc_clean(stripslashes($_GET['term']));
        } else {
            $term = wc_clean($term);
        }

        if (empty($term)) {
            die();
        }

        return $term;
    }

    /**
    * This function is used to fetch the details of variable product with corresponding variation.
    * @param object $variation_product variations product details.
    * @param int $variation_id Id of variation to be fetched
    * @param array $variation_data the variation combination of attributes.
    * @param array $found_products products found with similar data as searched
    * @param array $excluded_products product variation which are already included .
    * @return array $found_products variable product details of corresponding variation.
    */
    public static function addVariationInSearchResults($variation_product, $variation_id, $variation_data, $found_products, $excluded_products)
    {
        if (!is_array($variation_data)) {
            return $found_products;
        }

        // Return Products if value of  any variation attribute is empty
        foreach ($variation_data as $variation) {
            if (empty($variation)) {
                return $found_products;
            }
        }
        //Creating md5 hash to check whether current combinations of attributes already exists in the array or not.
        $variation_hash = md5($variation_id.'_'.implode('_', $variation_data));

        if (in_array($variation_hash, $excluded_products)) {
            return $found_products;
        }

        //Check if this variation already exists in the array
        if (!isset($found_products[$variation_hash])) {
            $img_url = self::getImgUrl($variation_id, $variation_product);
            $productID = self::getParentID($variation_product);
            $max_val =  0;
            $bundle_item_quantity = get_post_meta($variation_id, '_stock', true);
            if (!empty($bundle_item_quantity)) {
                $max_val = $bundle_item_quantity;
            }
            
            $sku = $variation_product->get_sku();
            $text_name = !empty($sku) ? self::getVariationName($variation_product)." (".$sku.")" : self::getVariationName($variation_product)." (#".$variation_id.")";
            $found_products[$variation_hash] = array(
                'product_id' => $productID,
                'parent_hash' => md5($productID),
                'variation_id' => $variation_id,
                'min_val' => 0,
                'max_val' => $max_val,
                'item_quantity' => $bundle_item_quantity,
                'bundle_quantity' => 0,
                'product_type' => 'variation',
                'price' => $variation_product->get_price(),
                'text_name' => $text_name,
                'variation_attributes' => $variation_data,
                'variation_string' => self::getVariations($variation_data, $variation_id),
                'product_title' => $variation_product->get_title(),
                'url' => admin_url("/post.php?post={$productID}&action=edit"),
                'permalink' => get_permalink($variation_id),
                'product_image' => $img_url,
                'sku' => $sku,
            );
        }

        return $found_products;
    }

    /**
    * Gets the Image URL of Product.
    * @param int $variation_id Variation Id
    * @param object $variation_product Variation of Product
    */
    public static function getImgUrl($variation_id, $variation_product)
    {
        if (isset($variation_id) && $variation_id != '') {
            $img_url = wp_get_attachment_url(get_post_thumbnail_id($variation_id));
        }
        if (!$img_url || $img_url == '') {
            $img_url = wp_get_attachment_url(get_post_thumbnail_id(self::getParentID($variation_product)));
        }
        if (!$img_url || $img_url == '') {
            $img_url = WC()->plugin_url().'/assets/images/placeholder.png';
        }
    }

    /**
    * Returns the Variations of Product in string format.
    * @param array $variation_data Variation data of Product.
    * @param int $variation_id Variation Id
    */
    public static function getVariations($variation_data, $variation_id)
    {
        // global $cpbFunctions;
        if (isset($variation_data) && $variation_data != '') {
            $variation_data = maybe_unserialize($variation_data);
            $variableProduct = wc_get_product($variation_id);
            $product_attributes = $variableProduct->get_attributes();
            return getCpbVariationString($variation_data, $variableProduct, $product_attributes);
        }

        return '';
    }

    /**
    * Get the Name of variation Product.
    * @param  object $variationProduct variation product object
    */
    public static function getVariationName($variationProduct)
    {

        if (version_compare(WC_VERSION, '3.0.0', '>=')) {
            return rawurldecode(self::generateVariationName($variationProduct));
        }
            
            return rawurldecode($variationProduct->get_formatted_name());
    }

    /**
     * Generates Variation Title for WC greater than 3.0
     *
     * This is the copy of WooCommerce's WC_Product_Variation_Data_Store_CPT::generate_product_title(). Because
     * direct call to this method is not possible as it is a protected method, we are creating a copy of it in plugin
     *
     * @param  object $product variation product object
     * @return string          title of variation
     */
    public static function generateVariationName($product)
    {
        $include_attribute_names = false;
        $attributes = (array) $product->get_attributes();

        // Determine whether to include attribute names through counting the number of one-word attribute values.
        $one_word_attributes = 0;
        foreach ($attributes as $name => $value) {
            if (false === strpos($value, '-')) {
                ++$one_word_attributes;
            }
            if ($one_word_attributes > 1) {
                $include_attribute_names = true;
                break;
            }
            unset($name);
        }

        $include_attribute_names = apply_filters('woocommerce_product_variation_title_include_attribute_names', $include_attribute_names, $product);
        $title_base_text         = get_post_field('post_title', $product->get_parent_id());
        $title_attributes_text   = wc_get_formatted_variation($product, true, $include_attribute_names);
        $separator               = ! empty($title_attributes_text) ? ' &ndash; ' : '';

        return apply_filters(
            'woocommerce_product_variation_title',
            $title_base_text . $separator . $title_attributes_text,
            $product,
            $title_base_text,
            $title_attributes_text
        );
    }

    /**
    * Get Parent product Id for variation id
    * @param object $variation_product Variation Id
    * @return int $productID  Parent product Id
    */
    public static function getParentID($variation_product)
    {
        if (version_compare(WC_VERSION, '3.0.0', '<')) {
            $productID = $variation_product->parent->id;
        } else {
            $productID = $variation_product->get_parent_id();
        }
        return $productID;
    }

    /**
    * This function is to fetch all variations of the variable Product in the search result.
    * It first gets all the attributes of the variable product which is similar to the product which is searched.
    * Then the variations are fetched according to the attributes combinations.
    * It is then checked that if variation with similar attribute combination is present in the variable product list or not.
    * If yes the variable product details for that variation is fetched and put in the search bar.
    * @param array $product product details
    * @param array $found_products product details of the products having similar term as searched.
    * @param array $excluded_products products which are included previously.
    * @return array $found_products products found with the required variations.
    */
    public static function addAllVariationsInSearchResults($post, $product, $found_products, $excluded_products = array())
    {
        /*
         * Below we will be calling get_variation_attributes function which makes a call to get_attributes.
         * get_variation_attributes uses name of attribute as a key in the returned array. We need to change the
         * behavior and we need to have a 'slug' as a key in the array returned by get_variation_attributes.
         * Therefore, we'll change the array returned by get_attributes. We'll keep name as slug in
         * get_attributes array
         */
        
        if (version_compare(WC_VERSION, '3.0.0', '<')) {
            add_filter('woocommerce_get_product_attributes', 'cpbModifyProductAttributeNames', 99, 1);
            $attributesArray = $product->get_variation_attributes();
            remove_filter('woocommerce_get_product_attributes', 'cpbModifyProductAttributeNames', 99, 1);
        } else {
            add_filter('woocommerce_product_get_attributes', 'cpbModifyProductAttributeNames', 99, 1);
            $attributesArray = $product->get_attributes();
            remove_filter('woocommerce_product_get_attributes', 'cpbModifyProductAttributeNames', 99, 1);
        }
        
        $found_products = self::getVariableProduct($post, $product, $found_products, $excluded_products);

        if (empty($attributesArray)) {
            return $found_products;
        }
        $setAttributeName = function ($value) {
            return 'attribute_'.$value;
        };

        //Sets attribute_ prefix to all keys of an array
        $attributesArray = array_combine(
            array_map($setAttributeName, array_keys($attributesArray)),
            $attributesArray
        );

        $variationCombinations = self::getArrayCombinations($attributesArray);
        if ($variationCombinations) {
            foreach ($variationCombinations as $singleVariationCombination) {
                if (version_compare(WC_VERSION, '3.0.0', '<')) {
                    $variation_id = $product->get_matching_variation($singleVariationCombination);
                } else {
                    $data_store   = \WC_Data_Store::load('product-variable');

                    $variation_id =  $data_store->find_matching_product_variation($product, $singleVariationCombination);
                }
                if ($variation_id) {
                    $variation_product = wc_get_product($variation_id);
                    if (version_compare(WC_VERSION, '3.0.0', '<')) {
                        $variation_product->variation_data = $singleVariationCombination;
                    } else {
                        //Setting Variation Attributes in WC 3.0 and greater
                        $variationData = array();
                        foreach ($singleVariationCombination as $key => $value) {
                            // Remove attribute prefix which meta gets stored with.
                            if (0 === strpos($key, 'attribute_')) {
                                $key = substr($key, 10);
                            }
                            $variationData[ $key ] = $value;
                        }
                        $variation_product->set_props(array('attributes' => $variationData));
                    }

                    $found_products = self::addVariationInSearchResults($variation_product, $variation_id, $singleVariationCombination, $found_products, $excluded_products);
                }
            }
        }
        return $found_products;
    }

    /**
    * This function gets the conditional query for product selection .
    * If there are any products in exclude i.e, products already included in Add-on list do not fetch them again.
    * If there is any limit specified for the no. of products , include the limit in query.
    * @return string $query conditional query.
    */
    public static function getQueryPart()
    {
        $query = '';
        if (!empty($_GET['exclude'])) {
            $query .= ' AND posts.ID NOT IN ('.implode(',', array_map('intval', explode(',', $_GET['exclude']))).')';
        }

        if (!empty($_GET['limit'])) {
            $query .= ' LIMIT '.intval($_GET['limit']);
        }

        return $query;
    }

    /**
    * This function gets the variations with the combinations of the attributes.
    * @param array $arrays array of the attributes.
    * @return array $result Variation combinations of all the attributes.
    */
    public static function getArrayCombinations($arrays)
    {
        $result = array(array());
        foreach ($arrays as $property => $property_values) {
            if (!is_array($property_values)) {
                   continue;
            }
            $tmp = array();
            foreach ($result as $result_item) {
                foreach ($property_values as $property_value) {
                    $tmp[] = array_merge($result_item, array($property => $property_value));
                }
            }
            $result = $tmp;
        }

        return $result;
    }

    /**
    * This function is used to get the excluded products i.e, the products which are already included previously in the Add-on list.
    * @return array $excludedProducts products previously included.
    */
    public static function getExcludedProducts()
    {
        $excludedProducts = array();

        if (!empty($_GET['exclude'])) {
            $excludedProducts = array_filter(array_unique(explode(',', $_GET['exclude'])));
        }
        return $excludedProducts;
    }
}
