<?php

namespace wisdmlabs\cpb;

/**
 * Product Bundle cart functions and filters.
 * Referred from Woocommerce Bundle Product plugin.
 */
// Exit if accessed directly
if (! defined('ABSPATH')) {
    exit;
}

class WdmWcProductCart
{
    public $wdm_addons_prefix = '';
    public $wdm_nyp_prefix    = '';

    /**
     * Setup cart class
     * Adds the action for the following functions for cart.
     * 1: Filter for Support of Product Addons Plugin
     * 2: Filter for Support of NYP Plugin
     * 3: Filter for Cart validation for products availability,etc.
     * 4: Filter to add bundled data to cart.
     * 5: Filter to distinguish Box-products from other simple products on cart
     * 6: Filter to distinguish Box-products from other simple products in order.
     * 7: Filter to get the data for the item in the cart for display on front-end.
     * 8: Action to add the custom box data in the order meta.
     * 9: Action to  Add bundled items to the cart
     * 10: Filter to Modify cart items for bundled pricing strategy
     * 11: Filter to Load bundle data from session into the cart
     * 12: Filter to Sync quantities of bundled items with bundle quantity
     * 13: Filter to remove the remove item link for the bundled products in cart.
     * 14: Filter to check if the cart is updated.
     * 15: Filter to validate the cart when it is updated.
     * 16: Action to change the quantities of the bundled product in the cart.
     * 17: Action to remove the bundled product when main CPB removed.
     * 18: Action to restore the bundled product when main CPB restored.
     * 19: Filter to Reinialize cart item data for re-ordering purchased orders.
     * 20: Filter to show the bundled items under CPB item of the cart widget.
     * 21: Filter to consider main CPB Product count in the cart-items count.
     */
    public function __construct()
    {
        //global $woocommerce, $wp_query;


        // Support for Product Addons Plugin
        add_filter('product_addons_field_prefix', array($this, 'wooBundlesAddonsCartPrefix'), 10, 2);

        // Support for NYP Plugin
        add_filter('nyp_field_prefix', array($this, 'woo_bundles_nyp_cart_prefix'), 10, 2);

        // Validate bundle add-to-cart
        add_filter('woocommerce_add_to_cart_validation', array($this, 'wooBundlesValidation'), 10, 6);

        // Add bundle-specific cart item data
        add_filter('woocommerce_add_cart_item_data', array($this, 'wooBundlesAddCartItemData'), 10, 2);
        
        //to distinguish Box-products from other simple products on cart & thank you page & order dashoard
        add_filter('woocommerce_cart_item_class', array($this, 'wdmCartItemClass'), 10, 3);
        add_filter('woocommerce_order_item_class', array($this, 'wdmOrderItemClass'), 10, 3);
        add_filter('woocommerce_get_item_data', array($this, 'wdmDisplayItemData'), 10, 2);
        if (version_compare(WC_VERSION, '3.0.0', '<')) {
            add_action('woocommerce_add_order_item_meta', array($this, 'wdmAddValuesToOrderItemMeta'), 1, 2);
        } else {
            add_action('woocommerce_new_order_item', array($this, 'wdmAddValuesToOrderItemMeta'), 1, 2);
        }

        // Add bundled items to the cart
        add_action('woocommerce_add_to_cart', array($this, 'wooBundlesAddBundleToCart'), 10, 6);

        // Modify cart items for bundled pricing strategy
        add_filter('woocommerce_add_cart_item', array($this, 'wooBundlesAddCartItemFilter'), 10, 2);
        add_filter('woocommerce_cart_item_price', array($this, 'hideAddonItemPrice'), 10, 3);
        add_filter( 'woocommerce_cart_item_subtotal', array($this, 'hideAddonItemPrice'), 10, 3);

        // Load bundle data from session into the cart
        add_filter('woocommerce_get_cart_item_from_session', array($this, 'wooBundlesGetCartDataFromSession'), 10, 2);

        // Sync quantities of bundled items with bundle quantity
        add_filter('woocommerce_cart_item_quantity', array($this, 'wooBundlesCartItemQuantity'), 10, 2);
        add_filter('woocommerce_cart_item_remove_link', array($this, 'wooBundlesCartItemRemoveLink'), 10, 2);

        // Sync quantities of bundled items with bundle quantity
        add_filter('woocommerce_update_cart_action_cart_updated', array($this, 'woocommerceUpdateCartActionCartUpdatedFunc'), 10, 1);
        add_filter('woocommerce_update_cart_validation', array($this, 'woocommerceUpdateCartValidationFunc'), 10, 4);
        add_action('woocommerce_after_cart_item_quantity_update', array($this, 'wooBundlesUpdateQuantityInCart'), 1, 2);
        add_action('woocommerce_before_cart_item_quantity_zero', array($this, 'wooBundlesUpdateQuantityInCart'), 1);
        add_action('woocommerce_cart_item_removed', array($this, 'wdmWooCartRemoveItemFunc'), 1, 2);
        add_action('woocommerce_cart_item_restored', array($this, 'wdmWooCartRestoreItemFunc'), 1, 2);

        // Put back cart item data to allow re-ordering of bundles
        add_filter('woocommerce_order_again_cart_item_data', array($this, 'wooBundlesOrderAgain'), 10, 3);

        add_action( 'woocommerce_before_calculate_totals', array($this, 'wdmBoxCalculateTotals'), 10, 1);

        // Filter cart widget items
        add_filter('woocommerce_widget_cart_item_visible', array($this, 'wooBundlesCartWidgetFilter'), 10, 3);

        // Filter cart item count
        add_filter('woocommerce_cart_contents_count', array($this, 'wooBundlesCartContentsCount'));
    }

    public function hideAddonItemPrice($price, $cart_item, $cart_item_key)
    {
        if (isset($cart_item['isProductCPBSubscription']) && !$cart_item['isProductCPBSubscription']) {
            return $price;
        }

        if (isset( WC()->cart->cart_contents[ $cart_item_key ]['wdm_custom_bundled_by'])) {
            return '';
        }

        return $price;
    }

    public function wdmBoxCalculateTotals($cart_object)
    {
        //  This is necessary for WC 3.0+
        // if (is_admin() && !defined( 'DOING_AJAX' ) ) {
        //     return;
        // } 
        foreach ($cart_object->get_cart() as $cart_item_key => $cart_item) {
            // child product price
            if (isset($cart_item['wdm_bundled_items']) && isset($cart_item['product_id']) && !isProductCPBSubscription($cart_item['product_id'])) {
                return;
            }

            if (isset( $cart_item['wdm_custom_bundled_by']) && $cart_item['wdm_custom_bundled_by'] != '') {
                $cart_item['data']->set_price(0);
            }
                $wdmBoxTotal = 0;
            // main product price
            if (isset($cart_item['wdm_bundled_items']) && (!empty($cart_item['wdm_bundled_items']))) {
                $wdmBoxAddonItems = $cart_item['wdm_custom_stamp'];
                $product_id  = $cart_item['product_id'];
                $pricing_type = get_post_meta($product_id, '_per_product_pricing_active', true);
                if ($pricing_type == 'yes') {
                    if (is_array($wdmBoxAddonItems) && count($wdmBoxAddonItems) > 0) {
                        foreach ($wdmBoxAddonItems as $addOnKey => $wdmAddonItem) {
                            $addonProduct = isset($wdmAddonItem['item_data']['variation_id']) ? wc_get_product($wdmAddonItem['item_data']['variation_id']) : wc_get_product($wdmAddonItem['item_data']['product_id']);
                            
                            // Removing this as on this filter the Subscription plugin is returning price as 0 if the product is non-subscription product
                            if (isProductCPBSubscription($product_id)) {
                                remove_filter('woocommerce_product_get_price', 'WC_Subscriptions_Cart::set_subscription_prices_for_calculation', 100, 2);
                                remove_filter('woocommerce_product_variation_get_price', 'WC_Subscriptions_Cart::set_subscription_prices_for_calculation', 100, 2);
                                
                                $displayPrice = wc_get_price_to_display($addonProduct);
                                add_filter('woocommerce_product_variation_get_price', 'WC_Subscriptions_Cart::set_subscription_prices_for_calculation', 100, 2);
                                add_filter( 'woocommerce_product_get_price', 'WC_Subscriptions_Cart::set_subscription_prices_for_calculation', 100, 2);
                            } else {
                                $displayPrice = wc_get_price_to_display($addonProduct);
                            }

                            $addonQty     = absint($wdmAddonItem['quantity'] ? $wdmAddonItem['quantity'] : 0);
                            if ( ! $wdmAddonItem || !in_array($wdmAddonItem['type'], array('simple', 'variation'))) {
                                continue;
                            }
                            $wdmBoxTotal += $displayPrice * $addonQty;
                        }
                    }
                } else {
                    $boxObject = wc_get_product($product_id);
                    $box_price = $cart_item['data']->price;
                    $wdmBoxTotal = $box_price;
                }

                // per item + base price
                $withBase = get_post_meta($product_id, '_product_base_pricing_active', true);
                if (($pricing_type == 'yes') && ( $withBase=='yes') && is_numeric($wdmBoxTotal)) {
                    $box_price = $cart_item['data']->price;
                    $wdmBoxTotal += $box_price;
                } 

                if (isset($cart_item['subscription_renewal']) && isset($cart_item['subscription_renewal']['line_item_id'])) {
                    $wdmBoxTotal = wc_get_order_item_meta($cart_item['subscription_renewal']['line_item_id'], '_line_total', true);
                }
                
                if (isset($cart_item['subscription_resubscribe']) && isset($cart_item['subscription_resubscribe']['line_item_id'])) {
                    $wdmBoxTotal = wc_get_order_item_meta($cart_item['subscription_resubscribe']['line_item_id'], '_line_total', true);
                }
                $cart_item['data']->set_price(floatval($wdmBoxTotal));
            }
        }        
    }

    /**
     * Sets a unique prefix for unique add-ons. The prefix is set and re-set
     * globally before validating and adding to cart.
     * @param string $class class for cart-item.
     * @param array $cart_item CPB cart-item
     * @return string $class with appended bundled class.
     */
    public function wdmCartItemClass($class, $cart_item)
    {
        if (isset($cart_item['wdm_custom_bundled_by'])) {
            return $class . " wdm_bundled_item";
        } else {
            return $class;
        }
    }
    /**
     * Sets a unique prefix for unique add-ons. The prefix is set and re-set
     * globally before validating and adding to order.
     * @param string $class class for order-item.
     * @param array $order_item CPB order-item
     * @return string $class with appended bundled class.
     */
    public function wdmOrderItemClass($class, $order_item)
    {
        if (isset($order_item['wdm_bundled_by'])) {
            return $class . " wdm_bundled_item";
        } else {
            return $class;
        }
    }

    /**
    * Gets the data for the item in the cart for display on front-end.
    * @param array $data single Item data.
    * @param array $cartItem cart-item.
    */
    public function wdmDisplayItemData($data, $cartItem)
    {
        global $woocommerce;

        $cart_contents = $woocommerce->cart->cart_contents;


//	    var_dump('$cart_contents');
	    //var_dump($cart_contents);
        if (isset($cartItem['wdm_custom_bundled_by']) && isset($cart_contents[$cartItem['wdm_custom_bundled_by']])) {
            $postObject = get_post($cart_contents[$cartItem['wdm_custom_bundled_by']]['data']->get_id());
            $cpb_name = $postObject->post_title ;
	        
	        $attribute_value = $cartItem['data']->get_attribute('pa_per-portion');
	        $measurement     = $cartItem['data']->get_attribute('pa_uom');
	        $piece_or_pieces = (int) $attribute_value > 1 ? ' pieces' : ' piece';
	        $measurement     = empty( $measurement ) ? $piece_or_pieces : $measurement;
	       
	        $data[] = array(
		        'key' => __('Portion', 'brewpixel-wc-custom-product-bundles'),
		        'value' => (int) $attribute_value . $measurement,
	        );
	        $data[] = array(
            'key' => __('Included with', 'brewpixel-wc-custom-product-bundles'),
            'value' => $cpb_name,
            );
        }
        return $data;
    }

    /**
    * Adds the CPB Item in cart to the order meta table.
    * @param int $item_id Order Item Id.
    * @param array $values values in the cart session for custom box data.
    */
    public function wdmAddValuesToOrderItemMeta($item_id, $values)
    {
        global $woocommerce;
        if (version_compare(WC_VERSION, '3.0.0', '<')) {
            if (isset($values['wdm_custom_bundled_by'])) {
                $postObject = get_post($woocommerce->cart->cart_contents[$values['wdm_custom_bundled_by']]['data']->get_id());
                $cpb_name = $postObject->post_title ;//['post']->post_title;
                wc_add_order_item_meta($item_id, __('Included with', 'brewpixel-wc-custom-product-bundles'), $cpb_name);
            }
        } else {
            if (isset($values->legacy_values['wdm_custom_bundled_by'])) {
                $postObject = get_post($woocommerce->cart->cart_contents[$values->legacy_values['wdm_custom_bundled_by']]['data']->get_id());
                $cpb_name = $postObject->post_title ;//['post']->post_title;
                wc_add_order_item_meta($item_id, __('Included with', 'brewpixel-wc-custom-product-bundles'), $cpb_name);
            }
        }
    }
    /**
    * Get the add-on prefix for products add-on plugin.
    * @param string $prefix product add-ons plugins prefix.
    * @return string prefix for Products add-on plugin.
    */
    public function wooBundlesAddonsCartPrefix($prefix)
    {
        if (! empty($this->wdm_addons_prefix)) {
            return $this->wdm_addons_prefix . '-';
        }

        return $prefix;
    }

    /**
     * Validates add-to-cart for bundles.
     * Check the various conditions for each bundled product of CPB to be add into 
     * cart , has_sold_individually, in stock, is purchasable etc.
     * @param bool $add to add or not into cart.
     * @param int $product_id Cart Product Id.
     * @param int $product_quantity quantity for the product.
     * @param int $variation_id Variation-Id for product
     * @param array $variations Variations data for product.
     * @param array $cart_item_data cart product data.
     */
    public function wooBundlesValidation(
        $add,
        $product_id,
        $product_quantity,
        $variation_id = '',
        $variations = array(),
        $cart_item_data = array()
    ) {
        if (! empty($_REQUEST['prefill_os_m']) && $_REQUEST['prefill_os_m'] == "disabled") {
            // wdm_pb_bundles_add_notice(sprintf(__('This Custom Box cannot be purchased as mandatory product #%d - %s has run out of stock.', 'brewpixel-wc-custom-product-bundles'), $product_id, get_the_title($product_id)), 'error');
            return false;
        }

        global $woocommerce;

        // Get product type
        $product_type = 'simple';
        $terms = get_the_terms($product_id, 'product_type');
        if (isset(current($terms)->name)) {
            $product_type = sanitize_title(current($terms)->name);
        }

        if (isset($variation_id) && !empty($variation_id)) {
            $produtObject = wc_get_product($variation_id);
            $product_type = $produtObject->get_type();
        }

         // prevent bundled items from getting validated - they will be added by the container item
        if (isset($cart_item_data[ 'is_bundled' ]) && isset($_GET[ 'order_again' ])) {
            return false;
        }
        if ($product_type == 'simple' || $product_type == 'variation') {
            $alreadyInCartFlag = false;
            $product = wc_get_product($product_id);
            $productID = isset($variation_id) && !empty($variation_id) ? md5($variation_id.'_'.implode('_', $variations)) : $product_id;
            $ProductInCart = $this->getCartItemQuantities();
            if (isset($ProductInCart[$productID]) && $ProductInCart[$productID] > 0) {
                $alreadyInCartFlag = true;
            }


            if ($alreadyInCartFlag) {
                $item_data = array(
                    'id' => $productID,
                    'text' => $product->get_title(),
                    'variation_id' => $variation_id,
                    'product_id' => $product_id,
                    'product_type' => $product_type
                    );
                return $this->bundledAddToCart($product, $product_id, $item_data, $product_quantity, $variation_id, $variations, $cart_item_data, false, true);
            }
        }

        if ($product_type == 'wdm_bundle_product') {
            $product = wc_get_product($product_id);

            if (! $product) {
                return false;
            }

            // check request and prepare variation stock check data
            $stock_check_data = array();

            // grab bundled items for that product.
            $wdm_bundled_items = $product->getWdmCustomBundledItems();

            if (empty($wdm_bundled_items)) {
                return $add;
            }
            foreach ($wdm_bundled_items as $bundled_item_id => $bundled_item) {
                $id                   = $bundled_item->item_data['id'];
                $prodId = $bundled_item->item_data['product_id'];
                $bundled_product_type = $bundled_item->item_data['product_type'];

                $item_quantity = $bundled_item->quantity;

                $quantity      = $item_quantity * $product_quantity;
                if ($bundled_product_type == 'variable') {
                    if (isset($cart_item_data[ 'wdm_custom_stamp' ][ $bundled_item_id ][ 'variation_id' ]) && isset($_GET[ 'order_again' ])) {
                        $variation_id = $cart_item_data[ 'wdm_custom_stamp' ][ $bundled_item_id ][ 'variation_id' ];
                    } else {
                        $variation_id = $_REQUEST[ 'bundle_variation_id' ][ $bundled_item_id ];
                    }

                    if (isset($variation_id) && $variation_id > 1) {
                        $stock_check_data[ $id ][ 'type' ] = 'variable';

                        $variation_stock = get_post_meta($variation_id, '_stock', true);

                        if (get_post_meta($variation_id, '_price', true) === '') {
                            wdm_pb_bundles_add_notice(sprintf(__('Sorry, the selected variation of &quot;%s&quot; cannot be purchased.', 'brewpixel-wc-custom-product-bundles'), get_the_title($id)), 'error');
                            return false;
                        }

                        if (! isset($stock_check_data[ $id ][ 'variations' ])) {
                            $stock_check_data[ $id ][ 'variations' ] = array();
                        }

                        if (! isset($stock_check_data[ $id ][ 'managed_quantities' ])) {
                            $stock_check_data[ $id ][ 'managed_quantities' ] = array();
                        }

                        if (! in_array($variation_id, $stock_check_data[ $id ][ 'variations' ])) {
                            $stock_check_data[ $id ][ 'variations' ][] = $variation_id;
                        }

                        // If stock is managed on a variation level
                        if (isset($variation_stock) && $variation_stock !== '') {
                            // If a stock-managed variation is added to the cart multiple times, its stock must be checked for the sum of all quantities
                            if (isset($stock_check_data[ $id ][ 'managed_quantities' ][ $variation_id ])) {
                                $stock_check_data[ $id ][ 'managed_quantities' ][ $variation_id ] += $quantity;
                            } else {
                                $stock_check_data[ $id ][ 'managed_quantities' ][ $variation_id ] = $quantity;
                            }
                        } else {
                            // Non-stock-managed variations of the same item
                            // must be stock-checked together
                            if (isset($stock_check_data[ $id ][ 'quantity' ])) {
                                $stock_check_data[ $id ][ 'quantity' ] += $quantity;
                            } else {
                                $stock_check_data[ $id ][ 'quantity' ] = $quantity;
                            }
                        }
                    } else {
                        wdm_pb_bundles_add_notice(__('Please choose product options&hellip;', 'brewpixel-wc-custom-product-bundles'), 'error');
                        return false;
                    }

                    // Verify all attributes for the variable product were set - TODO: verify with filters

                    $attributes = (array) maybe_unserialize(get_post_meta($id, '_product_attributes', true));
                    $variations = array();
                    $all_set    = true;

                    $variation_data = array();

                    $custom_fields = get_post_meta($variation_id);

                    // Get the variation attributes from meta
                    foreach ($custom_fields as $name => $value) {
                        if (! strstr($name, 'attribute_')) {
                            continue;
                        }

                        $variation_data[ $name ] = sanitize_title($value[ 0 ]);
                    }


                    // Verify all attributes
                    foreach ($attributes as $attribute) {
                        if (! $attribute[ 'is_variation' ]) {
                            continue;
                        }

                        $taxonomy = 'attribute_' . sanitize_title($attribute[ 'name' ]);

                        if (! empty($_REQUEST[ 'bundle_' . $taxonomy ][ $bundled_item_id ])) {
                            // Get value from post data
                            // Don't use woocommerce_clean as it destroys sanitized characters
                            $value = sanitize_title(trim(stripslashes($_REQUEST[ 'bundle_' . $taxonomy ][ $bundled_item_id ])));

                            // Get valid value from variation
                            $valid_value = $variation_data[ $taxonomy ];

                            // Allow if valid
                            if ($valid_value == '' || $valid_value == $value) {
                                continue;
                            }
                        } elseif (isset($cart_item_data[ 'wdm_custom_stamp' ][ $bundled_item_id ][ 'attributes' ]) && isset($cart_item_data[ 'wdm_custom_stamp' ][ $bundled_item_id ][ 'variation_id' ]) && isset($_GET[ 'order_again' ])) {
                            $value = sanitize_title(trim(stripslashes($cart_item_data[ 'wdm_custom_stamp' ][ $bundled_item_id ][ 'attributes' ][ esc_html($attribute[ 'name' ]) ]))); // $taxonomy in WC 2.1

                            $valid_value = $variation_data[ $taxonomy ];

                            if ($valid_value == '' || $valid_value == $value) {
                                continue;
                            }
                        }

                        $all_set = false;
                        wdm_pb_bundles_add_notice(__('Please choose product options&hellip;', 'brewpixel-wc-custom-product-bundles'), 'error');
                        return false;
                    }
                } elseif (in_array($bundled_product_type, array('simple', 'variation', 'subscription'))) {
                    $stock_check_data[ $id ][ 'type' ] = $bundled_product_type;
                    $stock_check_data[ $id ][ 'item_data' ] = $bundled_item->item_data;

                    if (isset($stock_check_data[ $id ][ 'quantity' ])) {
                        $stock_check_data[ $id ][ 'quantity' ] += $quantity;
                    } else {
                        $stock_check_data[ $id ][ 'quantity' ] = $quantity;
                    }
                }


                // Validate Product add-ons Plugin support
                global $Product_Addon_Cart;

                if (! empty($Product_Addon_Cart)) {
                    $this->wdm_addons_prefix = $prodId;

                    if (! $Product_Addon_Cart->validate_add_cart_item(true, $prodId, $quantity)) {
                        return false;
                    }

                    $this->wdm_addons_prefix = '';
                }

                // Validate nyp Plugin support

                if (get_post_meta($product_id, '_per_product_pricing_active', true) == 'yes' && function_exists('WC_Name_Your_Price')) {
                    $this->wdm_nyp_prefix = $prodId;

                    if (! WC_Name_Your_Price()->cart->validate_add_cart_item(true, $prodId, $quantity)) {
                        return false;
                    }

                    $this->wdm_nyp_prefix = '';
                }
            }


            // Check stock for bundled items one by one
            // If out of stock, don't proceed

            foreach ($stock_check_data as $item_id => $data) {
                if (in_array($data[ 'type' ], array('simple', 'variation', 'subscription')) && $data[ 'quantity' ] != 0) {
                    $currentVariationId = $data[ 'type' ] == "variation" ? $data['item_data']['variation_id']: '';
                    $varAttributes = $data[ 'type' ] == "variation" ? $data['item_data']['variation_attributes']: array();
                    $item_stock_status = get_post_meta($data['item_data']['product_id'], "_stock", true);
                    $single_product = wc_get_product($data['item_data']['product_id']);

                    //The most important condition to check if whether product is purchasable or not. All products in the box must be purchasable.
                    if (! $single_product->is_purchasable()) {
                        $this->addNoticeConditionally(false, 'is_pur', $product, $data['item_data']);
                        return false;
                    }
                    $addIN = $this->bundledAddToCart($product, $data['item_data']['product_id'], $data['item_data'], $data[ 'quantity' ], $currentVariationId, (array)$varAttributes, array(), false);
                    if (! $addIN) {
                        return false;
                    }
                }
            }
        }
        return $add;
    }

    /**
     * Adds bundle specific cart-item data.
     * Add the custom product box data to woocommerce cart session.
     * @param array $cart_item_data Cart Item Data.
     * @param int $product_id CPB Product Id.
     * @return array $cart_item_data bumdled products included in cart-item data.
     */
    public function wooBundlesAddCartItemData($cart_item_data, $product_id)
    {
        global $Product_Addon_Cart;

        // Get product type
        $terms        = get_the_terms($product_id, 'product_type');
        if (! empty($terms) && isset(current($terms)->name)) {
            $product_type = sanitize_title(current($terms)->name);
        } else {
            $product_type = 'simple';
        }

        if ($product_type == 'wdm_bundle_product') {
            $product = wc_get_product($product_id);

            if (! $product) {
                return false;
            }

            // grab bundled items
            $wdm_bundled_items = $product->getWdmCustomBundledItems();

            if (empty($wdm_bundled_items)) {
                return;
            }

            // Create a unique stamp id with the bundled items' configuration
            $stamp = array();

            foreach ($wdm_bundled_items as $bundled_item_id => $bundled_item) {
                $id                   = $bundled_item->product_id;
                $bundled_product_type = $bundled_item->product->get_type();
                $stamp[ $bundled_item_id ][ 'product_id' ] = $id;
                $stamp[ $bundled_item_id ][ 'wdm_cpb_id' ] = $product_id;
                $stamp[ $bundled_item_id ][ 'type' ]       = $bundled_product_type;
                $stamp[ $bundled_item_id ][ 'quantity' ]   = $bundled_item->quantity;
                $stamp[ $bundled_item_id ][ 'title' ]   = $bundled_item->item_data['text'];
                $stamp[ $bundled_item_id ][ 'item_data' ]   = $bundled_item->item_data;
                if ($bundled_item->item_data['product_type'] == "variation") {
                    $stamp[ $bundled_item_id ][ 'variation_attributes' ]   = $bundled_item->item_data['variation_attributes'];
                }

                // Store bundled item addons add-ons config in stamp to avoid generating the same bundle cart id
                if (! empty($Product_Addon_Cart)) {
                    $addon_data = array();

                    // Set addons prefix
                    $this->wdm_addons_prefix = $bundled_item_id;

                    $bundled_product_id           = $id;
                    // commented as unused
                    //$bundled_product_variation_id = isset($stamp[ $bundled_item_id ][ 'variation_id' ]) ? $stamp[ $bundled_item_id ][ 'variation_id' ] : '';

                    $addon_data = $Product_Addon_Cart->add_cart_item_data($addon_data, $bundled_product_id);

                    // Reset addons prefix
                    $this->wdm_addons_prefix = '';

                    if (! empty($addon_data[ 'addons' ])) {
                        $stamp[ $bundled_item_id ][ 'addons' ] = $addon_data[ 'addons' ];
                    }
                }
            }

            $cart_item_data[ 'wdm_custom_stamp' ] = $stamp;

            // Prepare additional data for later use
            $cart_item_data[ 'wdm_bundled_items' ] = array();

            return $cart_item_data;
        } else {
            return $cart_item_data;
        }
    }

    /**
    * Adds bundled items to the cart.
    * Gets the bundled products data from the woocommerce session and add into
    * cart if the cart-validation is successfull, recalculate the totals by 
    * adding into cart.
    * @param bool $bundle_cart_key CPB Product cart key.
    * @param int $bundle_id CPB Cart Product Id.
    * @param int $bundle_quantity quantity for the CPB product.
    * @param int $variation_id Variation-Id for CPB product
    * @param array $variations Variations data for CPB product.
    * @param array $cart_item_data cart CPB product data.
    */
    public function wooBundlesAddBundleToCart(
        $bundle_cart_key,
        $bundle_id,
        $bundle_quantity,
        $variation_id,
        $variation,
        $cart_item_data
    ) {
        global $woocommerce;
        if (isset($cart_item_data[ 'wdm_custom_stamp' ]) && ! isset($cart_item_data[ 'wdm_custom_bundled_by' ])) {
            $dynamic_pricing_allowed = get_post_meta($bundle_id, '_per_product_pricing_active', true);

            // this id is unique, so that bundled and non-bundled versions of the same product will be added separately to the cart.
            error_log("WHY :: ".isProductCPBSubscription($bundle_id));
            $bundleitem_cart_data = array('bundled_item_id' => '', 'bundle_cpb_id' => $bundle_id, 'wdm_custom_bundled_by' => $bundle_cart_key, 'wdm_custom_stamp' => $cart_item_data[ 'wdm_custom_stamp' ], 'dynamic_pricing_allowed' => $dynamic_pricing_allowed, 'isProductCPBSubscription' => isProductCPBSubscription($bundle_id));

            // the bundle
            $bundle = $woocommerce->cart->cart_contents[ $bundle_cart_key ][ 'data' ];
            // Now add all items - yay
            foreach ($cart_item_data[ 'wdm_custom_stamp' ] as $bundled_item_id => $bundled_item_stamp) {
                // identifier needed for fetching post meta
                $bundleitem_cart_data[ 'bundled_item_id' ] = $bundled_item_id;

                $item_quantity = $bundled_item_stamp[ 'quantity' ];
                $quantity      = $item_quantity * $bundle_quantity;

                $product_id = $bundled_item_stamp[ 'product_id' ];

                $bundled_product_type = $bundled_item_stamp[ 'type' ];

                if (in_array($bundled_product_type, array('simple'))) {
                    $variation_id = '';
                    $variations   = array();
                } elseif ($bundled_product_type == 'variation') {
                    $variation_id = $bundled_item_stamp[ 'product_id' ];
                    $variations   = (array)$bundled_item_stamp[ 'variation_attributes' ];
                }

                // Set addons and nyp prefix
                $this->wdm_addons_prefix = $this->wdm_nyp_prefix    = $bundled_item_id;

                // Add to cart
                $bundleitem_cart_key = $this->bundledAddToCart($bundle, $product_id, $bundled_item_stamp['item_data'], $quantity, $variation_id, $variations, $bundleitem_cart_data, true);

                if (!$bundleitem_cart_key) {
                    return false;
                } else {
                    if (! in_array($bundleitem_cart_key, $woocommerce->cart->cart_contents[ $bundle_cart_key ][ 'wdm_bundled_items' ])) {
                        $woocommerce->cart->cart_contents[ $bundle_cart_key ][ 'wdm_bundled_items' ][] = $bundleitem_cart_key;
                    }
                }
                
                // Reset addons and nyp prefix
                $this->wdm_addons_prefix = $this->wdm_nyp_prefix    = '';
            }
        }

        unset($bundle_id);
        unset($variation);
    }

    public function variationBundledAddToCart(
        $bundle,
        $product_id,
        $item_data,
        $qty_counted_in_cart,
        $quantity,
        $variation_id,
        $variation,
        $cart_item_data,
        $add,
        $cart_item_key,
        $cart_id,
        $fromWoo = false
    ) {
        global $woocommerce;
        $product_data = wc_get_product($variation_id ? $variation_id : $product_id);
        $parent_id = wp_get_post_parent_id($variation_id);
        $parentData = wc_get_product($parent_id);
        if ($product_data->backorders_allowed() || $parentData->backorders_allowed()) {
            // is_sold_individually
            if ($product_data->is_sold_individually() == 'yes' && $quantity > 1) {
                $this->addNoticeConditionally($add, 'isi', $bundle, $item_data);
                return false;
            }

            if ($parentData->is_sold_individually() == 'yes' && $quantity > 1) {
                $this->addNoticeConditionally($add, 'isi', $bundle, $item_data);
                return false;
            }
        } else {
             // Stock check - only check if we're managing stock and backorders are not allowed
            if (!$product_data->is_in_stock() && $quantity>0) {
                if (!$parentData->is_in_stock() && $quantity>0) {
                    $this->addNoticeConditionally($add, 'instk', $bundle, $item_data);
                    return false;
                }
                $this->addNoticeConditionally($add, 'instk', $bundle, $item_data);
                return false;
            }


            // is_sold_individually
            if (($product_data->is_sold_individually() == 'yes' || $parentData->is_sold_individually() == 'yes') && $quantity > 1) {
                $this->addNoticeConditionally($add, 'isi', $bundle, $item_data);
                return false;
            }

            if (! $product_data->has_enough_stock($quantity)) {
                if (! $parentData->has_enough_stock($quantity)) {
                    $this->addNoticeConditionally($add, 'hasstk', $bundle, $item_data);
                    return false;
                }
                $this->addNoticeConditionally($add, 'hasstk', $bundle, $item_data);
                return false;
            }
           
            // Stock check - this time accounting for whats already in-cart
            $product_qty_in_cart = $this->getCartItemQuantities();
            if ($product_qty_in_cart) {
                if ($parentData->is_sold_individually() == 'yes') {
                    if (isset($product_qty_in_cart[ $item_data['id'] ]) && ($product_qty_in_cart[ $item_data['id'] ] >= 1 && $quantity > 0)) {
                        if (! $add) {
                            $bundleTitle = $fromWoo ? "Cart" : $bundle->get_title();
                            wdm_pb_bundles_add_notice(sprintf(
                                '<a href="%s" class="button wc-forward">%s</a> %s',
                                wc_get_cart_url(),
                                __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                sprintf(__('You cannot add another &quot;%s&quot; to &quot;%s&quot; as it is already in your cart', 'brewpixel-wc-custom-product-bundles'), $item_data['text'], $bundleTitle)
                            ), 'error');
                        }
                        return false;
                    }
                } elseif ($product_data->managing_stock()) {
                    $managing_stock = get_post_meta($variation_id ? $variation_id : $product_id, '_manage_stock', true);
                    $qtyInCart = $this->getCartItemQuantitiesForStock();
                    if ($item_data['product_type'] == "variation" && $managing_stock == "no") {
                        $parent_id = wp_get_post_parent_id($variation_id);
                        if ($qtyInCart) {
                            if (isset($qtyInCart[ $parent_id ]) && ! $parentData->has_enough_stock($qtyInCart[ $parent_id ] - $qty_counted_in_cart + $quantity)) {
                                if (! $add) {
                                    if ($fromWoo) {
                                        $bundleTitle = $bundle->get_title();
                                        wdm_pb_bundles_add_notice(sprintf(
                                            '<a href="%s" class="button wc-forward">%s</a> %s',
                                            wc_get_cart_url(),
                                            __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                            sprintf(__('&quot;%s&quot; cannot be added to the cart because there is not enough stock of &quot;%s&quot; &mdash; we have %s in stock and you already have %s in your cart.', 'brewpixel-wc-custom-product-bundles'), $item_data['text'], $item_data['text'], $parentData->get_stock_quantity(), $qtyInCart[ $parent_id ])
                                        ), 'error');
                                    } else {
                                        $bundleTitle = $bundle->get_title();
                                        wdm_pb_bundles_add_notice(sprintf(
                                            '<a href="%s" class="button wc-forward">%s</a> %s',
                                            wc_get_cart_url(),
                                            __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                            sprintf(__('&quot;%s&quot; cannot be added to the cart because there is not enough stock of &quot;%s&quot; &mdash; we have %s in stock and you already have %s in your cart.', 'brewpixel-wc-custom-product-bundles'), $bundleTitle, $item_data['text'], $parentData->get_stock_quantity(), $qtyInCart[ $parent_id ])
                                        ), 'error');
                                    }
                                }
                                return false;
                            }
                        }
                    }

                    $managing_stock = $product_data->managing_stock();
                    // Variations
                    if ($variation_id && $managing_stock) {
                        if (isset($product_qty_in_cart[ $item_data['id'] ]) && ! $product_data->has_enough_stock($product_qty_in_cart[ $item_data['id'] ] - $qty_counted_in_cart + $quantity)) {
                            if (! $add) {
                                if ($fromWoo) {
                                    wdm_pb_bundles_add_notice(sprintf(
                                        '<a href="%s" class="button wc-forward">%s</a> %s',
                                        wc_get_cart_url(),
                                        __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                        sprintf(__('&quot;%s&quot; cannot be added to the cart because there is not enough stock of &quot;%s&quot; &mdash; we have %s in stock and you already have %s in your cart.', 'brewpixel-wc-custom-product-bundles'), $item_data['text'], $item_data['text'], $product_data->get_stock_quantity(), $product_qty_in_cart[ $item_data['id'] ])
                                    ), 'error');
                                } else {
                                    $bundleTitle = $bundle->get_title();
                                    wdm_pb_bundles_add_notice(sprintf(
                                        '<a href="%s" class="button wc-forward">%s</a> %s',
                                        wc_get_cart_url(),
                                        __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                        sprintf(__('&quot;%s&quot; cannot be added to the cart because there is not enough stock of &quot;%s&quot; &mdash; we have %s in stock and you already have %s in your cart.', 'brewpixel-wc-custom-product-bundles'), $bundleTitle, $item_data['text'], $product_data->get_stock_quantity(), $product_qty_in_cart[ $item_data['id'] ])
                                    ), 'error');
                                }
                            }
                            return false;
                        }
                    }
                }
            }
        }

        if (! $add) {
            return true;
        }

        // If cart_item_key is set, the item is already in the cart and its quantity will be handled by wooBundlesUpdateQuantityInCart.
        if (! $cart_item_key) {
            $cart_item_key = $cart_id;
            // Add item after merging with $cart_item_data - allow plugins and wooBundlesAddCartItemFilter to modify cart item
            
            $woocommerce->cart->cart_contents[ $cart_item_key ] = apply_filters('woocommerce_add_cart_item', array_merge($cart_item_data, array(
                'product_id'   => $item_data['product_id'],
                'variation_id' => $variation_id,
                'variation'    => $variation,
                'quantity'     => $quantity,
                'data'         => $product_data
            )), $cart_item_key);
        }
        return $cart_item_key;
    }

    /**
     * Add a bundled product to the cart. Must be done without updating session
     * data, recalculating totals or calling 'woocommerce_add_to_cart' recursively.
     * Check the various conditions for each bundled product of CPB to be add into * cart , has_sold_individually, in stock, is purchasable etc.
     * Provides the support with the Product-Add-ons and NYP Plugins.
     * @param object $bundle CPB Product.
     * @param int $product_id single product id.
     * @param int $quantity single product quantity.
     * @param int $variation_id variation-id for single product.
     * @param array $variation variation data for the single product.
     * @param array $cart_item_data cart-item-data.
     * @param bool $add whether to add in cart or not.
     */
    public function bundledAddToCart(
        $bundle,
        $product_id,
        $item_data,
        $quantity = 1,
        $variation_id = '',
        $variation = '',
        $cart_item_data = array(),
        $add = false,
        $fromWoo = false
    ) {
        global $woocommerce, $Product_Addon_Cart;
        if ($quantity < 0) {
            return false;
        }

        // Load cart item data when adding to cart
        if ($add) {
            // Add-ons cart item data is already stored in the composite_data array, so we can grab it from there.

            // Not doing so results in issues with file upload validation.
            if (! empty($Product_Addon_Cart) && isset($cart_item_data[ 'wdm_custom_stamp' ][ $cart_item_data[ 'bundled_item_id' ] ][ 'addons' ])) {
                remove_filter('woocommerce_add_cart_item_data', array($Product_Addon_Cart, 'add_cart_item_data'), 10, 2);

                $cart_item_data = (array) apply_filters('woocommerce_add_cart_item_data', $cart_item_data, $product_id, $variation_id);
                $cart_item_data[ 'addons' ] = $cart_item_data[ 'wdm_custom_stamp' ][ $cart_item_data[ 'bundled_item_id' ] ][ 'addons' ];

                add_filter('woocommerce_add_cart_item_data', array($Product_Addon_Cart, 'add_cart_item_data'), 10, 2);
            } else {
                $cart_item_data = (array) apply_filters('woocommerce_add_cart_item_data', $cart_item_data, $product_id, $variation_id);
            }

            // counted quantity
            $qty_counted_in_cart = $quantity;
        } else {
            $qty_counted_in_cart = 0;
        }

        // Generate a ID based on product ID, variation ID, variation data, and other cart item data
        $cart_id = $woocommerce->cart->generate_cart_id($product_id, $variation_id, $variation, $cart_item_data);

        // See if this product and its options is already in the cart
        $cart_item_key = $woocommerce->cart->find_product_in_cart($cart_id);

        if (!empty($variation_id)) {
            $parent_id = wp_get_post_parent_id($variation_id);
            $parentData = wc_get_product($parent_id);
            return $this->variationBundledAddToCart($bundle, $product_id, $item_data, $qty_counted_in_cart, $quantity, $variation_id, $variation, $cart_item_data, $add, $cart_item_key, $cart_id, $fromWoo);
        } else {
            // Ensure we don't add a variation to the cart directly by variation ID
            if ('product_variation' == get_post_type($product_id)) {
                $variation_id = $product_id;
                $product_id   = wp_get_post_parent_id($variation_id);
            }

            // Get the product
            if ($variation_id) {
                $product_data = wc_get_product($variation_id);
            } else {
                $product_data = wc_get_product($product_id);
            }

            if (! $product_data) {               
                return false;
            }
            
            // Check product is_purchasable
            if (! $product_data->is_purchasable()) {            
                $this->addNoticeConditionally($add, 'is_pur', $bundle, $item_data);
                return false;
            }

            //Check for backorder. If Backorder is allowed, check for sold_individually. If such sold_individually product is already in the cart, throw the error.
            if ($product_data->backorders_allowed()) {
                // is_sold_individually
                if ($product_data->is_sold_individually() == 'yes' && $quantity > 1) {
                    $this->addNoticeConditionally($add, 'isi', $bundle, $item_data);
                    return false;
                }
            } else {
                 // Stock check - only check if we're managing stock and backorders are not allowed
                if (! $product_data->is_in_stock() && $quantity>0) {
                    $this->addNoticeConditionally($add, 'instk', $bundle, $item_data);
                    return false;
                }

                // is_sold_individually
                if ($product_data->is_sold_individually() == 'yes' && $quantity > 1) {
                    $this->addNoticeConditionally($add, 'isi', $bundle, $item_data);
                    return false;
                } elseif (!empty($variation_id)) {
                    $parent_id = wp_get_post_parent_id($variation_id);
                    $variationParent = wc_get_product($parent_id);
                    if ($variationParent->is_sold_individually() == 'yes' && $quantity > 1) {
                        $this->addNoticeConditionally($add, 'isi', $bundle, $item_data);
                        return false;
                    }
                }

                if (! $product_data->has_enough_stock($quantity)) {
                    $this->addNoticeConditionally($add, 'hasstk', $bundle, $item_data);
                    return false;
                }
            // Stock check - this time accounting for whats already in-cart
                $product_qty_in_cart = $this->getCartItemQuantities();
                if ($product_data->managing_stock()) {
                    // Variations
                    if (!empty($variation_id) && $product_data->managing_stock()) {
                        if (isset($product_qty_in_cart[ $variation_id ]) && ! $product_data->has_enough_stock($product_qty_in_cart[ $variation_id ] - $qty_counted_in_cart + $quantity)) {
                            if (! $add) {
                                $bundleTitle = $fromWoo ? "Cart" : $bundle->get_title();
                                wdm_pb_bundles_add_notice(sprintf(
                                    '<a href="%s" class="button wc-forward">%s</a> %s',
                                    wc_get_cart_url(),
                                    __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                    sprintf(__('&quot;%s&quot; cannot be added to the cart because there is not enough stock of &quot;%s&quot; &mdash; we have %s in stock and you already have %s in your cart.', 'brewpixel-wc-custom-product-bundles'), $bundleTitle, $product_data->get_title(), $product_data->get_stock_quantity(), $product_qty_in_cart[ $variation_id ])
                                ), 'error');
                            }
                            return false;
                        }

                    // Products
                    } else {
                        if ($product_data->is_sold_individually() == 'yes') {
                            if (isset($product_qty_in_cart[ $product_id ]) && ($product_qty_in_cart[ $product_id ] >= 1 && $quantity > 0)) {
                                if (! $add) {
                                    $bundleTitle = $fromWoo ? "Cart" : $bundle->get_title();
                                    wdm_pb_bundles_add_notice(sprintf(
                                        '<a href="%s" class="button wc-forward">%s</a> %s',
                                        wc_get_cart_url(), //You cannot add another "Vneck Tshirt" to your cart.
                                        __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                        sprintf(__('You cannot add another &quot;%s&quot; to &quot;%s&quot; as it is already in your cart', 'brewpixel-wc-custom-product-bundles'), $item_data['text'], $bundleTitle)
                                    ), 'error');
                                }
                                return false;
                            }
                        } elseif (isset($product_qty_in_cart[ $product_id ]) && ! $product_data->has_enough_stock($product_qty_in_cart[ $product_id ] - $qty_counted_in_cart + $quantity)) {
                            if (! $add) {
                                $bundleTitle = $fromWoo ? "Cart" : $bundle->get_title();
                                wdm_pb_bundles_add_notice(sprintf(
                                    '<a href="%s" class="button wc-forward">%s</a> %s',
                                    wc_get_cart_url(),
                                    __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                    sprintf(__('&quot;%s&quot; cannot be added to the cart because there is not enough stock of &quot;%s&quot; &mdash; we have %s in stock and you already have %s in your cart.', 'brewpixel-wc-custom-product-bundles'), $bundleTitle, $product_data->get_title(), $product_data->get_stock_quantity(), $product_qty_in_cart[ $product_id ])
                                ), 'error');
                            }
                            return false;
                        }
                    }
                } elseif ($product_data->is_sold_individually() == 'yes') {
                    if (isset($product_qty_in_cart[ $product_id ]) && ($product_qty_in_cart[ $product_id ] >= 1 && $quantity > 0)) {
                        if (! $add) {
                            $bundleTitle = $fromWoo ? "Cart" : $bundle->get_title();
                            wdm_pb_bundles_add_notice(sprintf(
                                '<a href="%s" class="button wc-forward">%s</a> %s',
                                wc_get_cart_url(), //You cannot add another "Vneck Tshirt" to your cart.
                                __('View Cart', 'brewpixel-wc-custom-product-bundles'),
                                sprintf(__('You cannot add another &quot;%s&quot; to &quot;%s&quot; as it is already in your cart', 'brewpixel-wc-custom-product-bundles'), $item_data['text'], $bundleTitle)
                            ), 'error');
                        }
                        return false;
                    }
                }
            }

            if (! $add) {
                return true;
            }

            // If cart_item_key is set, the item is already in the cart and its quantity will be handled by wooBundlesUpdateQuantityInCart.
            if (! $cart_item_key) {
                $cart_item_key = $cart_id;
                // Add item after merging with $cart_item_data - allow plugins and wooBundlesAddCartItemFilter to modify cart item
                $woocommerce->cart->cart_contents[ $cart_item_key ] = apply_filters('woocommerce_add_cart_item', array_merge($cart_item_data, array(
                    'product_id'   => $product_id,
                    'variation_id' => $variation_id,
                    'variation'    => $variation,
                    'quantity'     => $quantity,
                    'data'         => $product_data
                )), $cart_item_key);
            }
            return $cart_item_key;
        }
        return $cart_item_key;
    }

    /**
     * Add notices conditionally on any error condition.
     * @param bool $add whether to add notice or not.
     * @param string $type Type of the notice.
     * @param object $bundle CPB Product.
     * @param object $product_data Single product in the bundle.
     */
    public function addNoticeConditionally($add, $type, $bundle, $item_data)
    {
        $product_data = wc_get_product(isset($item_data['variation_id']) && !empty($item_data['variation_id']) ? $item_data['variation_id'] : $item_data['product_id']);
        if (! $add) {
            if ($type == 'isi') {
                wdm_pb_bundles_add_notice(sprintf(__('&quot;%s&quot; cannot be added to the cart &mdash; only 1 &quot;%s&quot; can be purchased.', 'brewpixel-wc-custom-product-bundles'), $bundle->get_title(), $item_data['text']), 'error');
            } elseif ($type == 'is_pur') {
                wdm_pb_bundles_add_notice(sprintf(__('&quot;%s&quot; cannot be added to the cart because &quot;%s&quot; cannot be purchased.', 'brewpixel-wc-custom-product-bundles'), $bundle->get_title(), $item_data['text']), 'error');
            } elseif ($type == 'instk') {
                wdm_pb_bundles_add_notice(sprintf(__('&quot;%s&quot; cannot be added to the cart because &quot;%s&quot; is out of stock.', 'brewpixel-wc-custom-product-bundles'), $bundle->get_title(), $item_data['text']), 'error');
            } elseif ($type == 'hasstk') {
                wdm_pb_bundles_add_notice(sprintf(__('&quot;%s&quot; cannot be added to the cart because there is not enough stock of &quot;%s&quot; (%s remaining).', 'brewpixel-wc-custom-product-bundles'), $bundle->get_title(), $item_data['text'], $product_data->get_stock_quantity()), 'error');
            } elseif ($type == 'not_allowed') {
                wdm_pb_bundles_add_notice(sprintf(__('This Custom Box cannot be purchased as mandatory product #%d - %s has run out of stock.', 'brewpixel-wc-custom-product-bundles'), $bundle->get_id(), $item_data['text']), 'error');
            }
        }
    }

    /**
     * Change the product Shipping before add-to-cart.
     * @param array $cart_data Cart-item data.
     * @param int $itm_id addon Product Id.
     */
    public function wooBundlesAddCartItemFilter($cart_data, $itm_id)
    {
        global $woocommerce;
        $cart_contents = $woocommerce->cart->cart_contents;

        if (!isset($cart_data[ 'wdm_custom_bundled_by' ])) {
            return $cart_data;
        }

        $cpbBox = getCartAddonItemBox($cart_data);
        $addonId = $cart_data[ 'bundled_item_id' ];

        if($cpbBox && $cpbBox->hasAddon($addonId)) {
            $cart_data[ 'data' ]->set_virtual('yes');
            $cart_data[ 'data' ]->set_weight('');
            $cart_data[ 'data' ]->set_price(0);
        }

        return $cart_data;
    }

    /**
     * Reload all bundle-related session data in the cart.
     * Extract Custom Data from WooCommerce Session and Insert it into Cart Object.
     * @param array $cart_item Cart-Item data.
     * @param array $item_session_values cart session values.
     * @return array $cart_item cart-item object data for the cart object.
     */
    public function wooBundlesGetCartDataFromSession(
        $cart_item,
        $item_session_values
    ) {
        global $woocommerce;

        //removed as woocommerce 2.4.4 doesn't give cart_contents in get_cart()...Fatal Error: 5000 times repeadtedly called function

        $cart_contents = '';
        if (! empty($woocommerce->cart)) {
            $cart_contents = $woocommerce->cart->cart_contents;
        }
        if (isset($cart_item[ 'wdm_custom_bundled_by' ]) && !isset($cart_contents[$cart_item[ 'wdm_custom_bundled_by' ]])) {
            return $cart_item;
        }

        if (isset($item_session_values[ 'wdm_bundled_items' ])) {
            $cart_item[ 'wdm_bundled_items' ] = $item_session_values[ 'wdm_bundled_items' ];
        }

        if (isset($item_session_values[ 'wdm_custom_stamp' ])) {
            $cart_item[ 'wdm_custom_stamp' ] = $item_session_values[ 'wdm_custom_stamp' ];
        }

        if (isset($item_session_values[ 'wdm_custom_bundled_by' ])) {
            // load 'wdm_custom_bundled_by' field

            $cart_item[ 'wdm_custom_bundled_by' ] = $item_session_values[ 'wdm_custom_bundled_by' ];

            // load product bundle post meta identifier
            $cart_item[ 'bundled_item_id' ] = $item_session_values[ 'bundled_item_id' ];

            // load dynamic pricing permission
            $cart_item[ 'dynamic_pricing_allowed' ] = $item_session_values[ 'dynamic_pricing_allowed' ];

            // now modify item depending on bundle pricing & shipping options
            $bundle_cart_id = $cart_item[ 'wdm_custom_bundled_by' ];

            $per_product_pricing = false;
            if (isset($cart_contents[ $bundle_cart_id ][ 'data' ]->per_product_pricing_active)) {
                $per_product_pricing = $cart_contents[ $bundle_cart_id ][ 'data' ]->per_product_pricing_active;
            }

            $per_product_shipping = false;
            if (isset($cart_contents[ $bundle_cart_id ][ 'data' ]->per_product_shipping)) {
                $per_product_shipping = $cart_contents[ $bundle_cart_id ][ 'data' ]->per_product_shipping;
            }

            if ($per_product_pricing == false) {
                $cart_item[ 'data' ]->set_price(0);
                $cart_item[ 'data' ]->subscription_sign_up_fee = 0;
            }

            if (isset($cart_item[ 'wdm_custom_bundled_by' ])) {
                $cpbBox = getCartAddonItemBox($cart_item);
                $addonId = $cart_item[ 'bundled_item_id' ];
                if($cpbBox && $cpbBox->hasAddon($addonId)) {
                    $cart_item[ 'data' ]->set_virtual('yes');
                    $cart_item[ 'data' ]->set_weight('');
                }
            }
        }

        return $cart_item;
    }

    /**
     * Add "included with" bundle metadata.
     */
    public function wooBundlesGetItemData($data, $cart_item)
    {
        global $woocommerce;

        if (isset($cart_item[ 'wdm_custom_bundled_by' ]) && isset($cart_item[ 'wdm_custom_stamp' ])) {
            // not really necessary since we know its going to be there
            $product_key = $woocommerce->cart->find_product_in_cart($cart_item[ 'wdm_custom_bundled_by' ]);

            if (! empty($product_key)) {
                $product_name = get_post($woocommerce->cart->cart_contents[ $product_key ][ 'product_id' ])->post_title;
                $data[]       = array(
                    'name'    => __('Included with', 'brewpixel-wc-custom-product-bundles'),
                    'display' => __($product_name)
                );
            }
        }

        return $data;
    }

    /**
     * Bundled items can't be removed individually from the cart - this hides the * remove buttons.
     * @param string $link url or link when remove button clicked.
     * @param string $cart_item_key Cart-Item-Key
     */
    public function wooBundlesCartItemRemoveLink($link, $cart_item_key)
    {
        global $woocommerce;
        $cart_contents = $woocommerce->cart->cart_contents;

        if (!isset($cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ])) {
            return $link;
        }

        if (!isset($cart_contents[$cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ]]) && isset($cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ])) {
            WC()->cart->remove_cart_item($cart_item_key);
        }

        if (isset($cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ]) && isset($cart_contents[$cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ]])) {
            return '';
        }

        return $link;
    }

    /**
     * Bundled item quantities can't be changed individually. When adjusting
     * quantity for the container item, the bundled products must follow.
     * @param int $quantity cart-items quantity.
     * @param int quantity for the CPB Product in cart.
     */
    public function wooBundlesCartItemQuantity($quantity, $cart_item_key)
    {
        global $woocommerce;

        $cart_contents = $woocommerce->cart->cart_contents;

        if (!isset($cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ])) {
            return $quantity;
        }

        if (!isset($cart_contents[$cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ]]) && isset($cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ])) {
            WC()->cart->remove_cart_item($cart_item_key);
            return 0;
        }

        if (isset($woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_stamp' ])) {
            if (isset($woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ])) {
                return $woocommerce->cart->cart_contents[ $cart_item_key ][ 'quantity' ];
            }
        }

        return $quantity;
    }

    /**
     * Get cart items quantities - merged so we can do accurate stock checks on items across multiple lines.
     *
     * @return array
     */
    public function getCartItemQuantities() {
        global $woocommerce;
        $quantities = array();
        foreach ($woocommerce->cart->get_cart() as $cart_item_key => $values) {
            $product = $values['data'];
            $variationId = getProductId($product);
            $productID = $variationId;
            if ("product_variation" == get_post_type($variationId) && $values['variation_id'] == $variationId) {
                // $productObject = wc_get_product($variationId);
                $variations = $values['variation'];
                $variation_hash = md5($variationId.'_'.implode('_', $variations));
                // $variations = JsonProductSearch::getVariationProduct($variationId, $productObject, array(), array());
                $productID = $variation_hash;
            }
            $quantities[ $productID ] = isset($quantities[ $productID ]) ? $quantities[ $productID ] + $values['quantity'] : $values['quantity'];
        }

        return $quantities;
    }


    /*
     * Function returns qunatities in cart with the original product id and parent id for variations.
    */
    public function getCartItemQuantitiesForStock() {
        global $woocommerce;
        $quantities = array();
        foreach ($woocommerce->cart->get_cart() as $cart_item_key => $values) {
            $product = $values['data'];
            $variationId = getProductId($product);
            $productID = $variationId;
            $managing_stock = get_post_meta($variationId, '_manage_stock', true);
            if ("product_variation" == get_post_type($variationId) && $values['variation_id'] == $variationId && $managing_stock == "no") {
                $productID = wp_get_post_parent_id($variationId);
            }
            $quantities[ $productID ] = isset($quantities[ $productID ]) ? $quantities[ $productID ] + $values['quantity'] : $values['quantity'];
        }
        return $quantities;
    }

    /**
    * When the cart is updated, validate the various parameters for the product.
    * If the products in cart do not satisfy any of the conditions,display notice * messages.
    * @param boolean $update_status update status for the cart (true if cart
    * updated)
    * @param string $cart_item_key Cart-Item-Key
    * @param array $values values for cart-item.
    * @param int $quantity of the cart item.
    */
    public function woocommerceUpdateCartValidationFunc(
        $update_status,
        $cart_item_key,
        $values,
        $quantity
    ) {
        global $woocommerce;
        //check if the request is occurred only for custom product boxes.
        //If total quantity bundle quantity exceeds stock then add notice.
        if (isset($woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_stamp' ]) && ! empty($woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_stamp' ]) && !empty($quantity)) {
            //check if value is set then proceed to verify quantity exceeds or not.
            if (isset($values[ 'wdm_custom_stamp' ])) {
                $bundle_item_ids = $values[ 'wdm_custom_stamp' ];
                if (is_array($bundle_item_ids)) {
                    foreach ($bundle_item_ids as $item_id => $item_data) {
                        $itemObject = wc_get_product($item_data['product_id']);
                        if (isset($values[ 'data' ]->wdm_bundled_items) && !empty($item_data[ 'quantity' ])) {
                            $pre_item_data = $values[ 'data' ]->wdm_bundled_items;
                            $purchasable = $itemObject->is_purchasable();

                            if ($purchasable == 1) {
                                if (version_compare(WC_VERSION, '3.0.0', '<')) {
                                    $enable_stock = $pre_item_data[ $item_data['product_id'] ]->enable_stock;
                                    $wdm_custom_stkstatus = $itemObject->stock_status;
                                    $backordersAllowed = $itemObject->backorders;
                                    $soldIndividually = $itemObject->sold_individually;
                                } else {
                                    $enable_stock = $itemObject->managing_stock();
                                    $wdm_custom_stkstatus = $itemObject->get_stock_status();
                                    $backordersAllowed = $itemObject->get_backorders();
                                    $soldIndividually = $itemObject->is_sold_individually();
                                }

                                if ($soldIndividually == "yes") {
                                    $update_status = "false";
                                    wc_add_notice(__('Quantity exceeds the maximum quantity of certain products in the box. Kindly reduce the number of boxes selected.', 'brewpixel-wc-custom-product-bundles'), 'error');
                                    $_SESSION[ 'cart_update_status' ] = false;
                                    return false;
                                } elseif ($backordersAllowed == "no") {
                                    if ($wdm_custom_stkstatus == "out-of-stock") {
                                        $update_status = "false";
                                        wc_add_notice(__('Quantity exceeds the maximum quantity of certain products in the box. Kindly reduce the number of boxes selected.', 'brewpixel-wc-custom-product-bundles'), 'error');
                                        $_SESSION[ 'cart_update_status' ] = false;
                                        return false;
                                    } elseif ($enable_stock == "yes") {
                                        $bundle_item_max = $pre_item_data[ $item_data['item_data']['id'] ]->getMaxVal();
                                        if (! empty($bundle_item_max)) {
                                            if ($item_data[ 'quantity' ] * $quantity > $bundle_item_max) {
                                                $update_status = "false";
                                                wc_add_notice(__('Quantity exceeds the maximum quantity of certain products in the box. Kindly reduce the number of boxes selected.', 'brewpixel-wc-custom-product-bundles'), 'error');
                                                $_SESSION[ 'cart_update_status' ] = false;
                                                return false;
                                            }
                                        }
                                    }
                                } //
                            }
                        }//data
                    }//for
                }
            }
        }
        return $update_status;
    }

    /**
    * Returns true if cart is updated can be found through cart update status in
    * session variable.
    * @param boolean $cart_updated true if cart updated else false.
    * @return boolean $cart_updated true if cart updated else false.
    */
    public function woocommerceUpdateCartActionCartUpdatedFunc($cart_updated)
    {
        if (! empty($_SESSION)) {
            if (array_key_exists('cart_update_status', $_SESSION)) {
                if (! $_SESSION[ 'cart_update_status' ]) {
                    return false;
                }
            }
        }

        return $cart_updated;
    }

    /**
     * Keep quantities between bundled products and container items in sync.
     * Set the quantity for individual product in the bundle.
     * It will be multiplication of quantity of that product and quantity of the
     * main bundle.
     * @param string $cart_item_key cart-item-key.
     * @param int $quantity cart-item quantity.
     */
    public function wooBundlesUpdateQuantityInCart($cart_item_key, $quantity = 0)
    {
        global $woocommerce;

        if (isset($woocommerce->cart->cart_contents[ $cart_item_key ]) && ! empty($woocommerce->cart->cart_contents[ $cart_item_key ])) {
            if ($quantity == 0 || $quantity < 0) {
                $quantity = 0;
            } else {
                $quantity = $woocommerce->cart->cart_contents[ $cart_item_key ][ 'quantity' ];
            }

            if (isset($woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_stamp' ]) && ! empty($woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_stamp' ]) && ! isset($woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_bundled_by' ])) {
                // unique bundle wdm_custom_stamp added to all bundled items & the grouping item
                $stamp = $woocommerce->cart->cart_contents[ $cart_item_key ][ 'wdm_custom_stamp' ];

                // change the quantity of all bundled items that belong to the same bundle config
                foreach ($woocommerce->cart->cart_contents as $key => $value) {
                    if (isset($value[ 'wdm_custom_bundled_by' ]) && isset($value[ 'wdm_custom_stamp' ]) && $cart_item_key == $value[ 'wdm_custom_bundled_by' ] && $stamp == $value[ 'wdm_custom_stamp' ]) {
                        if ($value[ 'data' ]->is_sold_individually() && $quantity > 0) {
                            $woocommerce->cart->set_quantity($key, 1, false);
                        } else {
                            $bundle_quantity = $value[ 'wdm_custom_stamp' ][ $value[ 'bundled_item_id' ] ][ 'quantity' ];
                            $woocommerce->cart->set_quantity($key, $quantity * $bundle_quantity, false);
                        }
                    }
                }
            }
        }
    }

    /**
     * Remove child products when custom product box is removed from the cart.
     * @param string $cart_item_key cart-item-key.
     * @param array $cart-item cart-item data.
     */
    public function wdmWooCartRemoveItemFunc($cart_item_key, $cart_item)
    {
        global $woocommerce;
        if (isset($cart_item->cart_contents)) {
            foreach ($cart_item->cart_contents as $wdm_remove_key => $wdm_rmv_key_content) {
                if (isset($wdm_rmv_key_content[ 'wdm_custom_bundled_by' ])) {
                    if ($wdm_rmv_key_content[ 'wdm_custom_bundled_by' ] == $cart_item_key) {
                        $child_key = $wdm_remove_key;
                        if (! empty($child_key)) {
                            $woocommerce->cart->remove_cart_item($child_key);
                            unset($woocommerce->cart->cart_contents[ $child_key ]);
                        }
                    }
                }
            }
        }
    }

    /**
     * Restore child products when custom product box is restored from the cart.
     * When the undo option is clicked on the cart page,after clearing a cart-item
     * The single cart item can be restored.
     * @param string $cart_item_key cart-item-key.
     * @param array $cart-item cart-item data.
     */
    public function wdmWooCartRestoreItemFunc($cart_item_key, $cart_item)
    {
        global $woocommerce;

        foreach ($cart_item as $wdm_rmv_key_content) {
            if (is_array($wdm_rmv_key_content)) {
                foreach ($wdm_rmv_key_content as $child_key => $child_key_content) {
                    if (isset($child_key_content[ 'wdm_custom_bundled_by' ])) {
                        if ($child_key_content[ 'wdm_custom_bundled_by' ] == $cart_item_key) {
                            $woocommerce->cart->cart_contents[ $child_key ] = $child_key_content;
                            $woocommerce->cart->cart_contents[ $child_key ]['data'] = wc_get_product($child_key_content['variation_id'] ? $child_key_content['variation_id'] : $child_key_content['product_id']);
                        }
                    }
                }
            }
        }
    }

    /**
     * Reinialize cart item data for re-ordering purchased orders.
     * @param array $cart_item_data Cart-Item data.
     * @param array $order_item previous order item
     * @return array $cart_item_data Cart-Item data updated with the previous order
     */
    public function wooBundlesOrderAgain($cart_item_data, $order_item)
    {
        if (isset($order_item[ 'wdm_custom_bundled_by' ]) && isset($order_item[ 'wdm_custom_stamp' ])) {
            $cart_item_data[ 'is_bundled' ] = 'yes';
        }

        if (! isset($order_item[ 'wdm_custom_bundled_by' ]) && isset($order_item[ 'wdm_custom_stamp' ])) {
            $cart_item_data[ 'wdm_custom_stamp' ] = maybe_unserialize($order_item[ 'wdm_custom_stamp' ]);
        }

        return $cart_item_data;
    }

    /**
     * Do not show container items or bundled items, depending on the chosen
     * pricing method.
     * @param boolean $show to show cart-items or not.
     * @param array $cart_item Cart-Item data.
     * @return boolean $show to show or not.
     */
    public function wooBundlesCartWidgetFilter($show, $cart_item)
    {
        global $woocommerce;

        if (isset($cart_item[ 'wdm_custom_bundled_by' ])) {
            // not really necessary since we know its going to be there
            $bundle_key = $woocommerce->cart->find_product_in_cart($cart_item[ 'wdm_custom_bundled_by' ]);

            if (! empty($bundle_key)) {
                $product_id = $woocommerce->cart->cart_contents[ $bundle_key ][ 'product_id' ];

                if (get_post_meta($product_id, '_per_product_pricing_active', true) == 'no') {
                    return false;
                }
            }
        }

        if (! isset($cart_item[ 'wdm_custom_bundled_by' ]) && isset($cart_item[ 'wdm_custom_stamp' ])) {
            if (get_post_meta($cart_item[ 'product_id' ], '_per_product_pricing_active', true) == 'yes') {
                return false;
            }
        }

        return $show;
    }

    /**
     * Filters the reported number of cart items:
     * bundled items count is subtracted.
     * Only the main CPB Product count is considered
     * @param int $count cart-items count
     * @return int cart count when CPB product present.
     */
    public function wooBundlesCartContentsCount($count)
    {
        global $woocommerce;

        $cart = $woocommerce->cart->get_cart();

        $subtract = 0;
        foreach ($cart as $value) {
            if (isset($value[ 'wdm_custom_bundled_by' ])) {
                $bundle_cart_id    = $value[ 'wdm_custom_bundled_by' ];
                if (!isset($cart[ $bundle_cart_id ])) {
                    continue;
                }
                $bundle_product_id = $cart[ $bundle_cart_id ][ 'product_id' ];

                $per_product_shipping = (get_post_meta($bundle_product_id, '_per_product_shipping_active', true) == 'yes') ? true : false;

                if (! $per_product_shipping) {
                    $subtract += $value[ 'quantity' ];
                }
            }

            if (isset($value[ 'wdm_custom_stamp' ]) && ! isset($value[ 'wdm_custom_bundled_by' ])) {
                $bundle_product_id = $value[ 'product_id' ];

                $per_product_shipping = (get_post_meta($bundle_product_id, '_per_product_shipping_active', true) == 'yes') ? true : false;

                if ($per_product_shipping) {
                    $subtract += $value[ 'quantity' ];
                }
            }
        }
        return $count - $subtract;
    }
}
